import csv
from matplotlib import pyplot as plt
import bar_chart_race as bcr
import argparse
import sys
from licchart.helpers import catcher, dataperiod, datatodf, fetch, internetcheck, update, getjson, apichecker
import dateutil.parser
import regex as re

def licchart():
    parser = argparse.ArgumentParser(description='A Last.fm Bar Chart Race Maker.', epilog='Visit https://github.com/Stelvey/LICCHART/ for more info.')
    parser.add_argument('-v', '--version', help='print out current version and exit', action='version', version='LICCHART 0.1.0')

    parser.add_argument('source', help='str: your last.fm username or csv filename', nargs='?')

    parser.add_argument('-a', '--api', help='str: add/change custom api key and exit', metavar='KEY')

    group = parser.add_mutually_exclusive_group()
    group.add_argument('-m', '--months', help='takes less time to generate, but gives a less accurate result (default)', action='store_true')
    group.add_argument('-d', '--days', help='takes quite a while to generate, but gives very accurate result', action='store_true')

    parser.add_argument('-s', '--start', help='str (month first): set starting date (default: your first scrobble)', metavar='DATE')
    parser.add_argument('-e', '--end', help='str (month first): set ending date (default: your last scrobble)', metavar='DATE')

    parser.add_argument('-b', '--bars', type=int, default=20, help='int: set how many artists will be visible on the chart (default: %(default)s)', metavar='AMT')
    parser.add_argument('-l', '--length', type=float, help='float: set how long your animation will be (default: dynamic value)', metavar='MIN')
    parser.add_argument('-f', '--fps', type=int, default=30, help='int: more frames take more time to generate, but provide a smoother animation (default: %(default)s)', metavar='FPS')

    args = parser.parse_args()

    # Check internet
    if internetcheck():
        sys.exit(catcher(internetcheck()))

    # Some default chart values
    type = 'months'
    length = None

    # API input
    if args.api:
        rawjson = getjson({'api_key': args.api})
        if apichecker(rawjson) == 6:
            sys.exit(catcher(5))
        with open('api.key', 'w', encoding="utf-8") as f:
            f.write(args.api)
        return print('API key changed successfully')
    try:
        with open('api.key', 'r', encoding="utf-8") as f:
            api = f.read()
    except FileNotFoundError:
        print("You haven't set an API key!\nYou can get one at: https://www.last.fm/api/account/create\nPlease, enter it now:")
        api = input()
        if len(api) > 16:
            with open('api.key', 'w', encoding="utf-8") as f:
                f.write(api)
        else:
            sys.exit(catcher(5))

    # Type input
    if args.days:
        type = 'days'

    # Range inputs
    if args.strt:
        try:
            start = dateutil.parser.parse(args.strt).date()
        except ValueError:
            sys.exit(catcher(3))
    else:
        start = None
    if args.end:
        try:
            end = dateutil.parser.parse(args.end).date()
        except ValueError:
            sys.exit(catcher(3))
        if start > end:
            sys.exit(catcher(1))
    else:
        end = None

    # Bars, length, smooth inputs
    try:
        bars = int(args.bars)
    except ValueError:
        sys.exit(catcher(3))
    if bars < 1:
        sys.exit(catcher(3))

    if args.length:
        try:
            length = round(float(args.length), 2)
        except ValueError:
            sys.exit(catcher(3))
        if length < 0.1:
            sys.exit(catcher(3))

    try:
        smooth = int(args.fps) / 1000
    except ValueError:
        sys.exit(catcher(3))
    if smooth < 0.01:
        sys.exit(catcher(3))

    # Source input
    if args.source == None:
        print("You haven't specified your source (nickname or CSV filename).\nPlease, enter it now:")
        args.source = input()
        if not args.source:
            sys.exit(catcher(0))

    # Making a source var + support for " and '
    source = args.source.replace('[\"\'](.*)[\"\']', '')

    # This abomination finds CSV or uses username instead (to create/update data)
    try:
        file = open('LICCHART_' + source, 'r', encoding="utf-8")
    except FileNotFoundError:
        try:
            file = open('LICCHART_' + source + '.csv', 'r', encoding="utf-8")
        except FileNotFoundError:
            user = source
            data = fetch(user, api, 1, None)
        else:
            data = update(file, api)
            user = data[0][0].split('#', 1)[1]
    else:
        data = update(file, api)
        user = data[0][0].split('#', 1)[1]
    finally:
        if catcher(data) != None:
            sys.exit(catcher(data))

    # Making CSV file
    with open('LICCHART_' + user + '.csv', 'w', encoding="utf-8") as f:
        writer = csv.writer(f)
        writer.writerows(data)

    # Delete all scrobbles not in period range
    data = dataperiod(data, start, end)

    # Make DataFrame out of data list
    df = datatodf(data, type)
    if catcher(df) != None:
        sys.exit(catcher(df))

    # Update some default values
    if length:
        length = round(length * 60000 / len(df))
    else:
        if type == 'months':
            length = 1000
        else:
            length = 100
        
    smooth = round(smooth * length)
    print()
    print('         Periods: ' + str(len(df)))
    print('     Step length: ' + str(length))
    print('Steps per period: ' + str(smooth))
    print()
    print('Your animation is generating. It takes quite a while. Please, wait...')
    # Prepare a figure!
    fig, ax = plt.subplots(figsize = (7, 3.9), dpi = 144)
    # Chart background
    ax.set_facecolor('#1a0933')
    ax.tick_params(labelsize = 8, length = 0)
    # Set grid
    ax.grid(True, axis = 'x', color = '#32fbe2')
    # And make it below everything
    ax.set_axisbelow(True)
    # Remove black borders around chart
    [spine.set_visible(False) for spine in ax.spines.values()]
    # Video background
    fig.patch.set_facecolor('#1a0933')
    # Adjust width (fixes empty spaces)
    fig.subplots_adjust(left = 0.2, right = 0.95)
    # Set artists color
    ax.tick_params(colors='white', which='both')

    # Generate a chart!!!
    bcr.bar_chart_race(
        df = df,
        filename = 'LICCHART_' + user + '.mp4',
        fig = fig,
        shared_fontdict = {'color': 'white'},
        n_bars = bars,
        bar_label_size = 4,
        tick_label_size = 8,
        filter_column_colors = True,
        period_length = length,
        steps_per_period = smooth
    )

    print()
    print('#######################################################')
    print('#                                                     #')
    print('# Hey, check your folder. Your animation is complete! #')
    print('#                                                     #')
    print('#######################################################')
    print()


import cutlet
import regex as re
import pandas as pd
import requests
from datetime import datetime, timezone
from dateutil.relativedelta import relativedelta


# I'm very tired and I made this catcher for custom error codes
def catcher(code):
    codes = {
        0: "User not found",
        1: "Current range has no data",
        2: "This is not a correct data file",
        3: "Unsupported values",
        4: "API doesn't respond",
        5: "Invalid API key!\nPlease, set a valid API key with: --api\nYou can get one at: https://www.last.fm/api/account/create",
        6: "Connection failed :("
    }
    try:
        if code in codes:
            return 'Error: ' + codes[code]
    except:
        return None


# API-related errors
def apichecker(rawjson):
    try:
        # User not found error
        if rawjson['error'] == 6:
            return 0
        # API key error
        if rawjson['error'] == 10:
            return 6
    except KeyError:
        return None


# Internet connection checker
def internetcheck():
    try:
        apichecker(getjson(None))
    except requests.exceptions.ConnectionError:
        return 6
    else:
        return None


# Cut the now playing track, return cut version (it sometimes IndexErrors for some reason)
def cutnowplaying(scrobbles):
    try:
        if scrobbles[0].get('@attr'):
            return scrobbles[1:]
        else:
            return scrobbles
    except IndexError:
        print('IndexError occured (assuming user is not playing anything?)')


# Check if the file is correct
def iscsv(filename):
    return '.' in filename and filename.rsplit('.', 1)[1].lower() == 'csv'


# Get raw json from API call
def getjson(parameters):
    response = requests.get(f"https://ws.audioscrobbler.com/2.0/?method=user.getrecenttracks&format=json", params=parameters)
    return response.json()


# Return correct formatting (date or month)
def strftype(type):
    if type == 'months':
        return '%B %Y'
    elif 'days':
        return '%b %d, %Y'


# Turn timestamp into a date
def tstodate(ts):
        return datetime.fromtimestamp(int(ts), tz=timezone.utc).date()


# I have no idea what I'm doing (this is because datetime forces days)
def datecompare(date, type):
    if type == 'months':
        return [date.year, date.month]
    else:
        return [date.year, date.month, date.day]


# Prepare data list for the selected time period (not the fastest method)
def dataperiod(data, start, end):
    for scrobble in data[1:]:
        scrblts = tstodate(scrobble[0])
        if start and start > scrblts:
            data.remove(scrobble)
        if end and end < scrblts:
            data.remove(scrobble)
    return data


# Data to DF (days/months)
def datatodf(data, type):
    # Prepare dataframe, empty dict, loop iteration flag, day/month
    df = pd.DataFrame()
    dict = {}
    first = True

    # Cutlet var
    katsu = cutlet.Cutlet()

    print()
    print('Pushing scrobbles to DataFrame...')
    for scrobble in data[1:]:
        # Get current scrobble's date and artist
        scrbldate = tstodate(scrobble[0])
        # (These dollar signs were messing up the code, a temp fix?)
        scrblartist = scrobble[1].replace('$', 'S')
        # Turn "unknown glyphs" into romaji
        if re.match('[\p{Katakana}\p{Han}\p{Hiragana}\p{Bopomofo}]', scrblartist):
            scrblartist = katsu.romaji(scrblartist)
        # Get a date var the first while loop iteration
        if first:
            currentdate = scrbldate
            first = False
        
        # Check if it is time to push a day/month to dataframe
        if datecompare(currentdate, type) != datecompare(scrbldate, type):
            df = pd.concat([df, pd.DataFrame(dict, index=[currentdate.strftime(strftype(type))])])
            
            # This is for filling empty days/months (if there are any) with same data
            while datecompare(currentdate + relativedelta(**{type: 1}), type) < datecompare(scrbldate, type):
                currentdate += relativedelta(**{type: 1})
                df = pd.concat([df, pd.DataFrame(dict, index=[currentdate.strftime(strftype(type))])])

            # Update to a new date as well
            currentdate = scrbldate
        
        # Add a new artist to data or add a scrobble to an existing one
        if scrblartist in dict:
            dict[scrblartist] += 1.0
        else:
            dict[scrblartist] = 1.0

    # Push final day/month to dataframe (if any data was received)
    try:
        df = pd.concat([df, pd.DataFrame(dict, index=[currentdate.strftime(strftype(type))])])
    except UnboundLocalError:
        return 1

    # Fill in the NaN and give index a name ^^ owo
    df.fillna(0, inplace=True)
    df.index.name = 'Date'

    print('DataFrame is ready!')
    return df


# Collect all (optionally only new) scrobbles to data list!
def fetch(user, api, page, ts):
    # Set parameters for API call
    parameters = {
        'api_key': api,
        'limit': 1000,
        'user': user,
        'page': page
    }

    # Make an API call (checking it) and put JSON into a var
    rawjson = getjson(parameters)
    if apichecker(rawjson) != None:
        return apichecker(rawjson)

    # Call API again when it sometimes KeyErrors for some reason 
    success = False
    attempt = 0
    while not success:
        try:
            # Find out total pages
            pages = int(rawjson['recenttracks']['@attr']['totalPages'])
        except KeyError:
            print('KeyError occured, trying again')
            attempt += 1
            if attempt == 5:
                # API error
                return 4
            rawjson = getjson(parameters)
        else:
            success = True

    # Start with the last page (if not [UPDATING])
    if not ts:
        parameters['page'] = pages

    # Prepare data list with header
    data = [['timestamp#' + user, 'artist']]

    print()
    while True:
        # Stupidly make a 2nd API call, this time starting from the last page (and looping)
        rawjson = getjson(parameters)

        # Call API again when it sometimes KeyErrors for some reason
        success = False
        while not success:
            try:
                # Get all the scrobbles from a page
                scrobbles = rawjson['recenttracks']['track']
            except KeyError:
                print('KeyError occured, trying again')
                rawjson = getjson(parameters)
            else:
                success = True

        # Cut the now playing track
        scrobbles = cutnowplaying(scrobbles)

        for scrobble in reversed(scrobbles):
            # [UPDATING] Skip old scrobbles
            if ts and int(scrobble['date']['uts']) <= ts:
                continue
            # Fix "unknown date" scrobbles' dates
            elif int(scrobble['date']['uts']) <= 1108290000:
                scrbldate = 1108290000
            # Or just get current scrobble's date and artist
            else:
                scrbldate = scrobble['date']['uts']
            # (These dollar signs were messing up the code, a temp fix?)
            scrblartist = scrobble['artist']['#text'].replace('$', 'S')
            # Add to data list
            data += [[scrbldate, scrblartist]]
        
        # Switch to a new page
        print(f"FETCH: {pages - parameters['page'] + 1} out of {pages}")
        parameters['page'] -= 1

        if parameters['page'] < 1:
            break

    return data


# Update the CSV
def update(file, api):
    try:
        df = pd.read_csv(file)
    except:
        return 2

    # Get user
    user = df.columns[0].split('#', 1)[1]

    # Set parameters for API call
    parameters = {
        'api_key': api,
        'limit': 1000,
        'user': user,
        'page': 1
    }

    # Make an API call (checking it) and put JSON into a var
    rawjson = getjson(parameters)
    if apichecker(rawjson) != None:
        return apichecker(rawjson)

    # Call API again when it sometimes KeyErrors for some reason
    success = False
    attempt = 0
    while not success:
        try:
            # Find out total pages
            pages = int(rawjson['recenttracks']['@attr']['totalPages'])
        except KeyError:
            print('KeyError occured, trying again')
            attempt += 1
            if attempt == 5:
                return 0
            rawjson = getjson(parameters)
        else:
            success = True

    # Get latest timestamp
    ts = df.iloc[-1][0]

    # Find the page from where to start
    print()
    while True:
        # Call API again when it sometimes KeyErrors for some reason
        success = False
        while not success:
            try:
                # Get all the scrobbles from a page
                scrobbles = rawjson['recenttracks']['track']
            except KeyError:
                print('KeyError occured, trying again')
                rawjson = getjson(parameters)
            else:
                success = True

        # Cut the now playing track
        scrobbles = cutnowplaying(scrobbles)

        for scrobble in reversed(scrobbles):
            scrbldate = int(scrobble['date']['uts'])
            if scrbldate == ts:
                print('Pages collected! Starting your update')
                page = parameters['page']
                # Return old and now updated data (header + old + new w/o header)
                return [df.columns.values.tolist()] + df.values.tolist() + fetch(user, api, page, ts)[1:]
        
        # Switch to a new page
        print(f"Pages to update: {parameters['page']}")
        parameters['page'] += 1

        # In case that a user has nothing to do in life
        # And deletes that particular scrobble
        # Probably to fuck up the code
        if parameters['page'] > pages:
            return None

        # Make a new API call requesting next page
        rawjson = getjson(parameters)